/* DRAT.cpp 
   author Pieta Schofield
   date	 July 2007
   program to process the output from the TRiFLe.pl script
   search for combinations of enzymes that give a separation 
   the sequences into the assigned groups
   this code is all mine and if freely usable adaptable etc to anyone that
   wants it thought I would appriciate a credit
   
   This code when compiled is called from DRAT.pl perl script
   which uses the library from Tisdalls Beginning Perl for bioinformatics 
   for reading the REBASE files
*/
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <vector>
#include <utility>
#include <map>
#include <algorithm>
#include <stdlib.h>
#include <math.h>

#define VERSION "1.0.2 24-APR-2008"

using namespace std;

/*
* a class for the combinations and their scores (the output data)
*
* scores 0 - number of failures to resolve sequences in different groups
* 		 1 - number of 
*		 2 - number of within group differences
*/
class enzcomb{
public:
  enzcomb(){return;}
  void clear(){enzlist.clear();return;}
  void addenz(int eId){enzlist.push_back(eId);return;}
  void setscore(int iN, int fS){iScore[iN]=fS; return;}
  int getenz(int eId){return enzlist.at(eId);}
  vector<int> &getenzlist(){return enzlist;}
  float getscore(int iN){ return iScore[iN];}
  int size(){return enzlist.size();} 
  friend bool operator <(const enzcomb& comA, const enzcomb& comB);
private:
  vector<int> enzlist;
  int iScore[3];
};

bool operator <(const enzcomb& comA, const enzcomb& comB){
	if(comA.iScore[0]>comB.iScore[0])return true;
	if(comA.iScore[0]<comB.iScore[0])return false;
	if(comA.iScore[1]<comB.iScore[1])return true;
	if(comA.iScore[1]>comB.iScore[1])return false;
	if(comA.iScore[2]>comB.iScore[2])return true;
	return false;
}

// a class for the sequences and their cut sites (the input data)
class seqcut{
	public:
		seqcut(){return;}		
		void 	addcut(pair<int,int> npair){cuts.push_back(npair);return;}
		void 	setgid(int ngid){groupid=ngid;return;}
		void 	setlen(int nlen){seqlen=nlen;return;}
		void 	init(){cuts.clear();return;}
		void	setgname(string ngname){groupname = ngname; return;}
		void	setsname(string nsname){seqname = nsname; return;}
		int 	getgid(){return groupid;}
		int 	getlen(){return seqlen;}
		string	getgname(){return groupname;}
		string	getsname(){return seqname;}
		int 	getcut(int cutid, int endid);
		int 	dist(seqcut nseq);
		pair<int,int> getpair(int pid){return cuts[pid];}
		vector<pair<int,int> > getcuts(){return cuts;}
	private:
		string	groupname;
		string	seqname;
		int  	groupid;
		int		seqlen;
	    vector  < pair<int,int> > cuts;
};

int seqcut::getcut(int cutid, int endid){
	switch (endid){
	case 3:
		return cuts[cutid].second;
		break;
	case 5:
		return cuts[cutid].first;
		break;
	}		
}

// a routine to calculate the minimum end lengths
pair<int, int> minpair(pair<int, int> cA, pair<int, int> cB){
	int minfirst;
	int minsecond;
	minfirst=min(cA.first,cB.first);
	minsecond=min(cA.second,cB.second);
	return make_pair(minfirst,minsecond);
}

// a routine to calculate the distance between cuts
double pairdist(pair<int,int> cA, pair<int,int> cB){
	int fdist=cA.first-cB.first;
	int sdist=cA.second-cB.second;
	return sqrt((float)(fdist*fdist)+(float)(sdist*sdist));
}

// a routine to test if either manhattan distance is less than a threshhold
bool pairdistman(int mindist, pair<int, int> cA, pair<int, int> cB){
  if(abs(cA.first-cB.first)<mindist&&abs(cA.second-cB.second)<mindist)return true;
  return false;
}

// a routine to calculate the total number of combinations
long totalcombs(int n, int r){
	long c=1;
	if (r > n) return 0;
	for (long d=1; d <= r; d++){
		c *= n--;
		c /= d;
	}
	return c;
}

// Numerical Recipies like dynamic 2D-array creation routine
int **imatrix(int nrl,int nrh,int ncl,int nch)
{
  int i,**m;
  m=(int **)malloc((unsigned) (nrh-nrl+1)*sizeof(int*));
  m -= nrl;
  for(i=nrl;i<=nrh;i++) {
    m[i]=(int *)malloc((unsigned) (nch-ncl+1)*sizeof(int));
    m[i] -= ncl;
  }
  return m;
}

// Numerical Recipies like dynamic 2D-arry deletion routine
void free_imatrix(int **m,int nrl,int nrh,int ncl,int nch){
	int i;
	for(i=nrh;i>=nrl;i--) free((char*) (m[i]+ncl));
	free((char*) (m+nrl));
}

/* a routine (thread candidate or distributed process candidate) to score an enzyme combinations across all possible groups
*
*	iI - vector containing indexes of enzymes to test in combination
*	vD - The data
*	iT - Distance threshold
*	iF - Number of fails of worst of top List
*	pS - pointer to returned scores vector
*	iG - number of groups
*
*/
void threadCandidate(vector<int>& iI, vector<seqcut>& vD,  int iT, float iF, int * pS, int iG){
	int iPGid;
	int iFGid;
	int iTGid;
	pair<int, int> fpair;
	pair<int, int> ppair;
	int ** vInterGroupCount;
	int ** vInterGroupFail;
	if(iI.size()>1){
	  // if multiple enzyme digest make sure combination produces a unique cut
	  bool bloop=true;
	  int iSeq=0;
	  while (bloop){
		fpair=make_pair(vD[iSeq].getlen(),vD[iSeq].getlen());
		for(int owk=0;owk<iI.size();owk++){
		  fpair=minpair(fpair,vD[iSeq].getpair(iI[owk]));
		}
		bloop=false;
		for(int owk=0;owk<iI.size();owk++){
		  if(fpair==vD[iSeq].getpair(iI[owk]))bloop=true;
		}
		if(iSeq++==iI.size())bloop=false;
	  }
    }
	// create and initialize scores matrices
	int iScore=iG*(iG-1)/2; // assign initial score as perfect then count down on fails
 	vInterGroupFail=imatrix(0,iG,0,iG);
 	vInterGroupCount=imatrix(0,iG,0,iG);
	for(int gwk=0;gwk<iG;gwk++){
	  for(int hwk=0;hwk<iG;hwk++){
			vInterGroupFail[gwk][hwk]=0;
			vInterGroupCount[gwk][hwk]=0;
	  }
	}
	// This is the section that checks distance between each sequence for enzyme combination 
	for(int fwk=0;fwk<vD.size()-1;fwk++){
		iFGid=vD.at(fwk).getgid();
		for(int pwk=fwk;pwk<vD.size()-1;pwk++){
			iPGid=vD.at(pwk).getgid();
			if(iFGid>iPGid){
		  		iTGid=iFGid;
		  		iFGid=iPGid;
		  		iPGid=iTGid;
			}
			// this is where the metric generation goes
			fpair=make_pair(vD[fwk].getlen(),vD[fwk].getlen());
			ppair=make_pair(vD[pwk].getlen(),vD[pwk].getlen());
			for(int owk=0;owk<iI.size();owk++){
				fpair=minpair(fpair,vD[fwk].getpair(iI[owk]));
				ppair=minpair(ppair,vD[pwk].getpair(iI[owk]));
			}
			if(iFGid !=iPGid){
				if(pairdistman(iT,fpair,ppair)){
					if(vInterGroupFail[iPGid][iFGid]<1){
						iScore--;
					}
					vInterGroupFail[iPGid][iFGid]++;
		  	}
			}else{
		  	if(pairdistman(iT,fpair,ppair)){
			  	vInterGroupFail[iPGid][iFGid]++;  
		  	}
			}
			vInterGroupCount[iPGid][iFGid]++;
			if(iScore<iF){
				pS[0]=-1;
				free_imatrix(vInterGroupFail,0,iG,0,iG);
				free_imatrix(vInterGroupCount,0,iG,0,iG);
				return;
			}
		}
	}	
	/*
	   this is where the code to score the metric should go
	*/		
	int ifintra=0;
	int ifinter=0;
	for(int gwk=0;gwk<iG;gwk++){
	 	for(int hwk=0;hwk<iG;hwk++){
			if(vInterGroupCount[gwk][hwk]>0){
				if(hwk!=gwk)ifinter+=(int)(100.0*(float)vInterGroupFail[gwk][hwk]/(float)vInterGroupCount[gwk][hwk]);
				else ifintra+=(int)(100.0*(float)vInterGroupFail[gwk][hwk]/(float)vInterGroupCount[gwk][hwk]);
			}
		}
	}
	pS[0]=iScore;
	pS[1]=ifinter/(iG*(iG-1)/2);
	pS[2]=ifintra/iG;
	free_imatrix(vInterGroupFail,0,iG,0,iG);
	free_imatrix(vInterGroupCount,0,iG,0,iG);
	return;
}

/*
*
*	Routine to print out the scores and results files at the end of it all follows similar structure to scoring routine
*
*/
void printCandidate(string sF, vector<int>& iI, vector<seqcut>& vD, int iT, int iG, vector<string>& vG){
	int iPGid;
	int iFGid;
	int iTGid;
	pair<int, int> fpair;
	pair<int, int> ppair;
	int ** vInterGroupCount;
	int ** vInterGroupFail;
	vInterGroupFail=imatrix(0,iG,0,iG);
	vInterGroupCount=imatrix(0,iG,0,iG);
	ofstream outfile;
	outfile.open(sF.c_str());
	for(int gwk=0;gwk<iG;gwk++){
	  for(int hwk=0;hwk<iG;hwk++){
			vInterGroupFail[gwk][hwk]=0;
			vInterGroupCount[gwk][hwk]=0;
	  }
	}
	for(int fwk=0;fwk<vD.size()-1;fwk++){
		iFGid=vD.at(fwk).getgid();
		for(int pwk=fwk;pwk<vD.size()-1;pwk++){
			iPGid=vD.at(pwk).getgid();
			if(iFGid>iPGid){
		  iTGid=iFGid;
		  iFGid=iPGid;
		  iPGid=iTGid;
		} 
		// this is where the metric generation goes				
	   fpair=make_pair(vD[fwk].getlen(),vD[fwk].getlen());
		 ppair=make_pair(vD[pwk].getlen(),vD[pwk].getlen());
		 for(int owk=0;owk<iI.size();owk++){
			 fpair=minpair(fpair,vD[fwk].getpair(iI[owk]));
			  ppair=minpair(ppair,vD[pwk].getpair(iI[owk]));
		  }
		  if(iFGid !=iPGid){
			  if(pairdistman(iT,fpair,ppair)){
					vInterGroupFail[iPGid][iFGid]++;
			  }
			}else{
			  if(pairdistman(iT,fpair,ppair)){
				  vInterGroupFail[iPGid][iFGid]++;  
			  }
			}
			vInterGroupCount[iPGid][iFGid]++;
		}
	}
//	   this is where the code to score the metric should go		
//	cout << sF << endl;
	int ogwk;
	int ohwk;
	int igwk;
	int ihwk;
	outfile << "group,";
	for(ogwk=0;ogwk<iG-1;ogwk++)outfile << vG[ogwk] << ",";
	outfile << vG[ogwk]<< endl;	
	for(ogwk=0;ogwk<iG;ogwk++){
		outfile << vG[ogwk] << ",";
	  	for(ohwk=0;ohwk<iG-1;ohwk++){
	  		if(ogwk<ohwk){
				igwk=ohwk;
				ihwk=ogwk;
			}else{
				igwk=ogwk;
				ihwk=ohwk;
			}	  		
		  	if(vInterGroupFail[igwk][ihwk]>0){
				if(ihwk!=igwk)outfile << "X, ";
				else outfile << (int)(100*(vInterGroupFail[igwk][ihwk])/(float)(vInterGroupCount[igwk][ihwk])) << ",";
		  	}else {
				outfile  << "0,";
		  	}
	  	}
		if(vInterGroupFail[ogwk][ohwk]>0){
			if(ohwk!=ogwk)outfile << "X\n";
			else outfile << (int)(100*(vInterGroupFail[ogwk][ohwk])/(float)(vInterGroupCount[ogwk][ohwk])) << "\n";
		}else {
			outfile  << "0\n";
		}	  
	}
	outfile.close();
	free_imatrix(vInterGroupFail,0,iG,0,iG);
	free_imatrix(vInterGroupCount,0,iG,0,iG);
	return;
}

/* 
*
* the main code reads the input file then generates all combinations of enzymes for test finally outputs results
*
*
*/
int main(int argc, char* argv[]){
	int ewk;
	int gwk;
	int hwk;
	int iSeqCount=0;
	int iPos=0;
	int iFieldCount=0;
	int iLineCount=0;
	int iCutCount=0;
	int iGroupCount=0;
	string ch;
	string ch5;
	string end_line("END");
	ifstream infile;
	ofstream outfile;
	vector<seqcut> vInData;
	vector<string> vEnzymename;
	vector<string> vGroupname;
	vector<int> indx;
	vector<enzcomb> vGooduns;
	map<string, int> mGroupname;
	pair<int, int> fpair;
	seqcut newseq;
	/*
	*  Open and read in the file produced by the perl script
	*  Most of this next section reads in the enzyme names and then
	*  churns through the file creating a usable structure
	*  First output version no for user to ensure correct match of programs
	*
	*	Then get the command line parameters
	*	argv[1] - filename of cut data file produced by PERL script
	*	argv[2] - max number of enzyme per digest
	*	argv[3] - distance threshold for rejection as same
	*	argv[4] - number of enzymes to keep in the top N list
	*
	*/
	cout << "Score program :" << VERSION << endl;
	string filename(argv[1]);
	int k=atoi(argv[2]);
	int dThresh=atoi(argv[3]);
	int iTop=atoi(argv[4]);
	string infilename=filename+".out\0";
	string outfilename=filename+"_cuts.csv\0";
	string scorefilename=filename+"_scores.csv\0";
	cout << "Reading " << infilename << endl;
	infile.open(infilename.c_str());
	if(!infile){
		cout << "file not found \n";
		return 0;
	}
	while(!infile.eof()){
		// get the first line with the column names and form a look up of enzymes names vs columns
		if(iLineCount<1){
			infile >> ch;
			if(ch==end_line){
				iPos=0;
				iLineCount++;
				iFieldCount=0;
			}else{
				if(iFieldCount>2){
					if(iPos<1){
						iPos=1;
					}else{
						iPos=0;
						iCutCount++;
						vEnzymename.push_back(ch.substr(0,ch.length()-2));
					}
				}
				iFieldCount++;		
			}
		}else{
		// get the subsequent lines with the group id, sequence length and 5'- 3' pairs t
			infile >> ch;
			if(ch==end_line){
				iSeqCount++;
				vInData.push_back(newseq);
				iPos=0;
				iLineCount++;
				iFieldCount=0;
			}else{
				switch (iFieldCount){
				case 0:
					newseq.init();
					if(mGroupname.find(ch)==mGroupname.end()){
						mGroupname[ch]=iGroupCount++;
						vGroupname.push_back(ch);
					}
					newseq.setgid(mGroupname.find(ch)->second);
					newseq.setgname(ch);
					break;
				case 1:
					newseq.setsname(ch);
					break;
				case 2:
					newseq.setlen(atoi(ch.c_str()));
					break;
				default:
					if(iPos<1){
						iPos=1;
						ch5=ch;
					}else{
						iPos=0;
						newseq.addcut(make_pair(atoi(ch5.c_str()), atoi(ch.c_str())));
					}
				}
				iFieldCount++;
			}
		}
	}
	infile.close();
	cout << "processing" << endl;
	/*
	  This is the business end of the program generates all unique n-enzyme combinations and test each in
	  turn checking of group locations are unique and if that calculates the distances for each enzyme 
	*/
	if(k<1)k=1;
	int iFailCount;
	float iFail=-1;
	int combslistlength=0;
	int  nwk;
	int qwk;
	if(dThresh<1)dThresh=1;
	int n=vEnzymename.size();
	bool bFound=false;
	int pScore[3];
	enzcomb vEnzComb;
	/*
	*	Loop through first single enzymes then combinations of 2 then 3 etc until max enzymes in digest k is reached
	*/
	for(int twk=1;twk<=k;twk++){
		int r=twk;
		int total=totalcombs(n,r);
		bool done=true;
		for(int iwk=0;iwk<r;iwk++)indx.push_back(iwk);
		while(done){
			bool bReject=false;
			done=false;
			iFailCount=0;
			vEnzComb.clear();
			for(int owk=0;owk<r;owk++){
				vEnzComb.addenz(indx[owk]);
			}
			if(vGooduns.size()<iTop){
			  iFail=-1;
			}else{
			  iFail=vGooduns[iTop-1].getscore(0);
			}
			/*
			*	Check the combination of enzymes in vector indx
			*/
			threadCandidate(indx, vInData, dThresh, iFail, pScore, mGroupname.size());
			/*
			* Check returned score if necessary ammend scores vector drop last and add new
			*/
			for(int swk=0;swk<3;swk++)vEnzComb.setscore(swk,pScore[swk]);
			if(vGooduns.size()<iTop){
				vGooduns.push_back(vEnzComb);
			}else{
				if(vGooduns[iTop-1].getscore(0)<=pScore[0]){
					if(vGooduns[iTop-1].getscore(0)<pScore[0]){
						vGooduns[iTop-1]=vEnzComb;
					}else if (vGooduns[iTop-1].getscore(1)>=pScore[1]){
						if(vGooduns[iTop-1].getscore(1)>pScore[1]){
							vGooduns[iTop-1]=vEnzComb;
						}else if(vGooduns[iTop-1].getscore(2)<pScore[2]){
							vGooduns[vGooduns.size()-1]=vEnzComb;
						}
					}
				}
			}
			/*
			*	Resort the scores vector
			*/
			sort(vGooduns.begin(),vGooduns.end());
			/*
			*
			* The next loop generates the next set of enzymes to be tested
			*
			*/
			for(int iwk=r-1;iwk>=0;iwk--){
				if(indx[iwk]<=(n-1)-(r-iwk)){
					indx[iwk]++;
					for(int swk=iwk+1;swk<r;swk++){
						indx[swk]=indx[swk-1]+1;
					}
					iwk=-1;
					done=true;
				}	
			}
		}	
		indx.clear();
	}
	/* 
	   Write out the top N ranking enzyme sequences
	*/
	if(vGooduns.size()<=0)return 0;
	cout << "Writing " << outfilename << endl;
	outfile.open(outfilename.c_str());
	outfile << "Group,";
	outfile << "GID,";
	outfile << "Seq,";
	outfile << "Length,";
	for(nwk=0;nwk<vGooduns.size()-1;nwk++){	 
		for(ewk=0;ewk<vGooduns[nwk].size()-1;ewk++){
		  outfile << vEnzymename[vGooduns[nwk].getenz(ewk)] <<"_";
		}
		outfile << vEnzymename[vGooduns[nwk].getenz(ewk)] << "_5,";
		for(ewk=0;ewk<vGooduns[nwk].size()-1;ewk++){
		  outfile << vEnzymename[vGooduns[nwk].getenz(ewk)] <<"_";
		}
		outfile << vEnzymename[vGooduns[nwk].getenz(ewk)] << "_3,";
	}
	for(ewk=0;ewk<vGooduns[nwk].size()-1;ewk++){
	  outfile << vEnzymename[vGooduns[nwk].getenz(ewk)] <<"_";
	}
	outfile << vEnzymename[vGooduns[nwk].getenz(ewk)] << "_5,";
	for(ewk=0;ewk<vGooduns[nwk].size()-1;ewk++){
	  outfile << vEnzymename[vGooduns[nwk].getenz(ewk)] <<"_";
	}
	outfile << vEnzymename[vGooduns[nwk].getenz(ewk)] << "_3\n";
   	for(int fwk=0;fwk<iSeqCount-1;fwk++){
		outfile << vInData[fwk].getgname() << ",";
		outfile << vInData[fwk].getgid() << ",";
		outfile << vInData[fwk].getsname() << ",";
		outfile << vInData[fwk].getlen() << ",";
		int fixwk;
		for(fixwk=0;fixwk<vGooduns.size()-1;fixwk++){
			fpair=make_pair(vInData[fwk].getlen(),vInData[fwk].getlen());
			for(ewk=0;ewk<vGooduns[fixwk].size();ewk++){
			  fpair=minpair(fpair,vInData[fwk].getpair(vGooduns[fixwk].getenz(ewk)));
			}
			outfile << fpair.first << "," << fpair.second << ",";
		}
			fpair=make_pair(vInData[fwk].getlen(),vInData[fwk].getlen());
			for(ewk=0;ewk<vGooduns[fixwk].size();ewk++){
			  fpair=minpair(fpair,vInData[fwk].getpair(vGooduns[fixwk].getenz(ewk)));
			}
			//		outfile << 3 << "," << 3 ;
				outfile << fpair.first << "," << fpair.second ;
		outfile << endl;
	}
	outfile.close();
	//
	outfile.open(scorefilename.c_str());
	outfile << "Enzyme name(s), total group combs, group combs success, Ave % seq fails, Ave group fidelity" << endl; 
	for(qwk=0;qwk<vGooduns.size();qwk++){
		string enzfilename=filename+"_";
		for(ewk=0;ewk<vGooduns[qwk].size()-1;ewk++){
		  enzfilename.append(vEnzymename[vGooduns[qwk].getenz(ewk)]);
			enzfilename.append("_");
		} 
		enzfilename.append(vEnzymename[vGooduns[qwk].getenz(ewk)]);
		enzfilename.append(".csv");
		outfile << enzfilename << "," << (iGroupCount)*(iGroupCount-1)/2 << "," ;
		for(int swk=0;swk<2;swk++) outfile << vGooduns[qwk].getscore(swk) << " , ";
		outfile <<   vGooduns[qwk].getscore(2)<<endl;
		printCandidate(enzfilename, vGooduns[qwk].getenzlist(), vInData, dThresh, mGroupname.size(), vGroupname);
	}
	outfile.close();
  	cout << "DONE" << endl;
	return 0;

}

