#!/usr/bin/perl
# 
my $version=" VERSION -  1.0.2 24-APR-2008";
#
# Simple Perl script to run through a directory of fasta files and calculate the terminal
# restriction fragment lengths for all the restriction enzymes in the rebase database.
# Expects to find the sequences to be grouped in fasta files, one file per group in a directory
# below the one in which the script is run. Output files have the same name as the directory
#
# Code borrows heavily from Tisdall's Perl for Biolimformatics books and uses Tisdall 
# library funtions to read REBASE
#
# author Pieta Schofield 2007
# Dundee University / SCRI
#
# amended to run from commandline params
# amended to search either or both directions 
#
use lib "./lib/";
use strict;
use warnings;
use File::Basename;
use BeginPerlBioinfo;  
use Getopt::Long;
# Declare variables
my ($fasta, $maxenz, $mindist, $topn, $sense, $enzfile, $enzname );
usage() if ( ! GetOptions('fasta=s' => \$fasta, 'maxenz=i' => \$maxenz, 'mindist=i' => \$mindist, 'topn=i' => \$topn, 'sense=s' =>\$sense, 'enzfile=s' => \$enzfile, 'enzname=s'=> \$enzname));
my %seq_hash=( );
my $sequence;
my $writesequence=0;
my $name;
my $line;
#check the command line
print "$version\n";  
chomp($fasta);
# loop through each file making sure it has .fasta extention and read data into
# a hash table using the fasta file name as the group name for the sequences within
my $fastafile=$fasta . ".fasta";
open(FASTAFILE, "<$fastafile");
while(<FASTAFILE>){
	# Read in the file 
	$line=$_;
	# discard blank line
	if ($line =~ /^\s*$/) {
		next;
# discard comment line
	} elsif($line =~ /^\s*#/){ 
	# discard fasta header line
		next;
	} elsif($line =~ /^>/) {
		$line  =~ s/\s//g;
		if($writesequence){
			$name =~ s/>//g;
			my $groupname = substr($name,0,3);
			$sequence =~ s/-//g;
			my $revsequence=reverse($sequence);
			$revsequence=~ tr/ACGTacgt/TGCAtgca/;
			$sequence .= " $revsequence";
			$sequence .= " $groupname";
			$seq_hash{$name}=$sequence;
		}
		$writesequence=1;
		$name = $line;
		$sequence = "";
		# keep line, add to sequence string
	} else {
		$line  =~ s/\s//g;
		$sequence .= "$line";
	}
}
if($writesequence){
	$name =~ s/>//g;
	my $groupname = substr($name,0,3);
	$sequence =~ s/-//g;
	my $revsequence=reverse($sequence);
	$revsequence=~ tr/ACGTacgt/TGCAtgca/;
	$sequence .= " $revsequence";
	$sequence .= " $groupname";
	$seq_hash{$name}=$sequence;
}

my $seqfile=$fasta .  "_fasta.csv";
unless(open(OPD,">$seqfile")){
	print "Cannot create output";
	exit;
}
for $sequence (sort keys %seq_hash){
	my @seqfields = split(" ",$seq_hash{$sequence});
	print OPD "$sequence $seqfields[2] $seqfields[0] $seqfields[1] END\n";
}
close(OPD);
close(FASTAFILE);
%seq_hash=();
# Declare variables
chomp($enzfile);
my %name_hash = (  );
open(REBASE, "<$enzfile");
# Read in the REBASE file this is pretty much unchanged from Tisdall
while(<REBASE>) {
#	Discard header lines
	( 1 .. /Rich Roberts/ ) and next;
#	Discard blank lines
	/^\s*$/ and next;    
# 	Split the two (or three if includes parenthesized name) fields
	my @fields = split( " ", $_);
	my $cart;
	my $cartpos;
 	my $adj;
# 	Get and store the recognition site
	my $site = pop @fields;
	my $entry = $fields[0];
	my $regexp= IUB_to_regexp($site);
	if(not defined $name_hash{$entry}) {
		$cart=$site;
		$cart=~/\^/ig;
		$cartpos= (length($cart) - pos($cart));
		$name_hash{$entry} = "$regexp $cartpos";
	}else{
		$cart=$site;
		$cart=~/\^/ig;
		$cartpos= (length($cart) - pos($cart));
		$name_hash{$entry} .= " $regexp $cartpos";
	}
}
close(REBASE);

my %isohash=();
my $refile=$fasta . "_iso.csv";
unless(open(OPD,">$refile")){
	print "Cannot create output";
	exit;
}

for my $name (sort keys %name_hash){
	if(not defined $isohash{$name_hash{$name}}){
		$isohash{$name_hash{$name}}=$name;
		my @fields=split(" ",$name_hash{$name});
		my $cutsites=scalar(@fields)/2;
		print OPD "$name $cutsites $name_hash{$name} END\n";
}
}
close(OPD);
%isohash=();
%name_hash=();

my %regexps=();
my %adjusts=();
open(REFILE, "<$refile");
while(<REFILE>){
	my $rline=$_;
	my @rfields=split(" ",$rline);
	my $count=$rfields[1];
	my $num=1;
	$regexps{$rfields[0]}="$rfields[2]";
	$adjusts{$rfields[0]}="$rfields[3]";
	while($num<$count){
		$num=$num+1;
		my $renum=(2*$num);
		my $ajnum=(2*$num) +1;
		$regexps{$rfields[0]}.=" $rfields[$renum]";
		$adjusts{$rfields[0]}.=" $rfields[$ajnum]";
	}
}
close(REFILE);

my $outfile=$fasta . ".out";
open(OUTFILE, ">$outfile");
print OUTFILE "Group Sequence Length ";
for my $re (sort keys %regexps){
	if($re eq $enzname or $enzname eq "all"){
		print OUTFILE $re ."_5 ". $re . "_3 ";
	}
}
print OUTFILE "END\n";
open(FASTAFILE, "<$seqfile");
while(<FASTAFILE>){
	my $fline=$_;
	my @ffields=split(" ",$fline);
	my $seqlen=length($ffields[2]);
	print OUTFILE "$ffields[1] $ffields[0] $seqlen ";
	for my $re (sort keys %regexps){
		my @refield="";
		my @ajfield="";
		my $regex="";
		if($re eq $enzname or $enzname eq "all"){
			my $i3min=$seqlen;
			my $i5min=$seqlen;
		   @refield=split(" ",$regexps{$re});
			my $cutsites=scalar(@refield);
			@ajfield=split(" ",$adjusts{$re});
			my $expno=0;
			while ($expno<$cutsites){
				my $forseq=$ffields[2];
				my $revseq=$ffields[3];
				if($sense eq "r"){
					$forseq="";
				}
				if ($sense eq "f"){
					$revseq="";
				}
				my $i3=$seqlen;
				my $i5=$seqlen;
			    $regex=$refield[$expno];
				my $adj=$ajfield[$expno];
				$regex=~s/\s//g; 
				if($forseq=~ m/($regex)/ig){
					$i5 = pos($forseq) - $adj +1;
				}
				if($revseq=~ m/($regex)/ig){
					$i3 = pos($revseq) -$adj +1;
				}
				if($i5<$i5min){
					$i5min=$i5;
				}
				if($i3<$i3min){
					$i3min=$i3;
				}
				$expno=$expno+1;
			}
			print OUTFILE "$i5min $i3min ";
		}
	}
	print OUTFILE "END\n";
}
close(FASTAFILE);
close(OUTFILE);

# run the scoring program
my @args=("./DRAT", "$fasta", "$maxenz", "$mindist", "$topn");
system(@args)==0 or die "system @args failed: $?";
exit;

sub usage
{
	print "Unknown option: @_\n" if ( @_ );
	print "usage: program -fasta  -maxenz -mindist -top -enzfile -sense \n";
	exit;
}


